from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Sequence

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Expression
    from tumfl.Token import Token


class Block(Statement):
    """A block consisting of statements"""

    def __init__(
        self,
        token: Token,
        statements: Sequence[Statement],
        returns: Optional[Sequence[Expression]],
    ):
        super().__init__(token, "Block")
        self.statements: list[Statement] = list(statements)
        self.returns: Optional[list[Expression]] = (
            list(returns) if returns is not None else None
        )



import unittest

from tumfl import parse
from tumfl.AST import Assign, Chunk
from tumfl.to_python_type import NilVal, Retype, ToPythonType

# TODO: Do full test by testing against interpreter, which in turn is tested against lua tests


class BaseClass(unittest.TestCase):
    def run_test(self, code: str, expected: Retype):
        ast = parse(f"a={code}")
        assert isinstance(ast, Chunk)
        first_node = ast.statements[0]
        assert isinstance(first_node, Assign)
        result = ToPythonType()(first_node.expressions[0])
        self.assertEqual(result, expected)
        self.assertIsInstance(result, type(expected))


class TestBoolean(BaseClass):
    def test_true(self):
        self.run_test("true", True)

    def test_false(self):
        self.run_test("false", False)


class TestNil(BaseClass):
    def test_nil(self):
        self.run_test("nil", NilVal())


class TestNumber(BaseClass):
    def test_integer(self):
        self.run_test("1", 1)

    def test_float(self):
        self.run_test("1.0", 1.0)
        self.run_test("1.23", 1.23)

    def test_hexadecimal(self):
        self.run_test("0x1", 1)
        self.run_test("0x1.23", 1.13671875)
        self.run_test("0x1.23p-4", 0.071044921875)
        self.run_test("0x1.23p4", 18.1875)


class TestString(BaseClass):
    def test_string(self):
        self.run_test('"hello"', "hello")


class TestTable(BaseClass):
    def test_key_table(self):
        self.run_test("{a = 1, b = 2}", {"a": 1, "b": 2})

    def test_not_defined(self):
        self.run_test("{a = 1, b = b}", None)

    def test_list(self):
        self.run_test("{1, 2, 3}", [1, 2, 3])
        self.run_test("{1, 2, a = 3}", {1: 1, 2: 2, "a": 3})
        self.run_test("{1, 2, 3, [4] = 4}", [1, 2, 3, 4])
        self.run_test("{1, 2, [5] = 5}", {1: 1, 2: 2, 5: 5})
        self.run_test("{[0] = 1, 1, 2}", {0: 1, 1: 1, 2: 2})

    def test_explicit_field(self):
        self.run_test('{["foo"] = 1, [2] = 2}', {"foo": 1, 2: 2})
        self.run_test("{[a] = 1}", None)


class TestUnOp(BaseClass):
    def test_not(self):
        self.run_test("not true", False)
        self.run_test("not false", True)
        self.run_test("not nil", True)
        self.run_test("not 1", False)
        self.run_test("not 0", False)
        self.run_test('not ""', False)
        self.run_test('not "hello"', False)

    def test_minus(self):
        self.run_test("-1", -1)
        self.run_test("-1.0", -1.0)
        self.run_test("-0x1", -1)
        self.run_test("-{1}", None)

    def test_bit_not(self):
        self.run_test("~1", -2)
        self.run_test("~0x1", -2)
        self.run_test("~0x1.23", -2)
        self.run_test("~{1}", None)

    def test_hash(self):
        self.run_test("#1", None)
        self.run_test('#"hello"', 5)
        self.run_test("#{1, 2, 3}", 3)
        self.run_test("#{a = 1, b = 2}", 2)
        self.run_test("#{[a] = 1}", None)


class TestBinOp(BaseClass):
    def test_invalid(self):
        self.run_test("a + b", None)

    def test_plus(self):
        self.run_test("1 + 2", 3)
        self.run_test("1.0 + 2.0", 3.0)
        self.run_test('"hello" + " world"', None)
        self.run_test("{1} + {2}", None)

    def test_minus(self):
        self.run_test("1 - 2", -1)
        self.run_test("1.0 - 2.0", -1.0)
        self.run_test('"hello" - " world"', None)
        self.run_test("{1} - {2}", None)

    def test_mult(self):
        self.run_test("1 * 2", 2)
        self.run_test("1.0 * 2.0", 2.0)
        self.run_test('"hello" * " world"', None)

    def test_div(self):
        self.run_test("1 / 2", 0.5)
        self.run_test("1.0 / 2.0", 0.5)
        self.run_test("1 / 1", 1.0)
        self.run_test('"hello" / " world"', None)

    def test_mod(self):
        self.run_test("1 % 2", 1)
        self.run_test("1.0 % 2.0", 1.0)
        self.run_test('"hello" % " world"', None)

    def test_integer_div(self):
        self.run_test("1 // 2", 0)
        self.run_test("2 // 2.0", 1)
        self.run_test('"hello" // " world"', None)

    def test_power(self):
        self.run_test("2 ^ 3", 8.0)
        self.run_test("2.0 ^ 3.0", 8.0)
        self.run_test("2 ^ 3.0", 8.0)
        self.run_test('"hello" ^ " world"', None)

    def test_concat(self):
        self.run_test('"hello" .. " world"', "hello world")
        self.run_test("{1} .. {2}", None)
        self.run_test("1 .. 2", "12")
        self.run_test('1 .. " hello"', "1 hello")

    def test_bit_and(self):
        self.run_test("1 & 2", 0)
        self.run_test("1.0 & 2.0", 0)
        self.run_test("1.5 & 2.0", None)
        self.run_test("1 & 3", 1)
        self.run_test('"hello" & " world"', None)

    def test_bit_or(self):
        self.run_test("1 | 2", 3)
        self.run_test("1.0 | 2.0", 3)
        self.run_test("1.5 | 2.0", None)
        self.run_test("1 | 3", 3)
        self.run_test('"hello" | " world"', None)

    def test_bit_xor(self):
        self.run_test("1 ~ 2", 3)
        self.run_test("1.0 ~ 2.0", 3)
        self.run_test("1.0 ~ 2.5", None)
        self.run_test("1 ~ 3", 2)
        self.run_test('"hello" ~ " world"', None)

    def test_shift_left(self):
        self.run_test("1 << 2", 4)
        self.run_test("1.0 << 2.0", 4)
        self.run_test("1.5 << 2.0", None)
        self.run_test('"hello" << " world"', None)

    def test_shift_right(self):
        self.run_test("1 >> 2", 0)
        self.run_test("4 >> 2", 1)
        self.run_test("1.0 >> 2.0", 0)
        self.run_test("1.5 >> 2.0", None)
        self.run_test('"hello" >> " world"', None)

    def test_equals(self):
        self.run_test("1 == 2", False)
        self.run_test("1.0 == 2.0", False)
        self.run_test('"hello" == " world"', False)
        # objects are compared by id by default
        self.run_test("{1} == {2}", None)
        self.run_test("1 == 1", True)
        self.run_test('"hello" == "hello"', True)

    def test_not_equals(self):
        self.run_test("1 ~= 2", True)
        self.run_test("1.0 ~= 2.0", True)
        self.run_test('"hello" ~= " world"', True)
        # objects are compared by id by default
        self.run_test("{1} ~= {2}", None)
        self.run_test("1 ~= 1", False)
        self.run_test('"hello" ~= "hello"', False)

    def test_greater_than(self):
        self.run_test("1 > 2", False)
        self.run_test("1.0 > 2.0", False)
        self.run_test('"hello" > "world"', False)
        self.run_test("{1} > {2}", None)
        self.run_test("1 > 1", False)
        self.run_test('"hello" > "hello"', False)
        self.run_test('"hello" > 1', None)

    def test_greater_equals(self):
        self.run_test("1 >= 2", False)
        self.run_test("1.0 >= 2.0", False)
        self.run_test('"hello" >= "world"', False)
        self.run_test("{1} >= {2}", None)
        self.run_test("1 >= 1", True)
        self.run_test('"hello" >= "hello"', True)
        self.run_test('"hello" >= 1', None)

    def test_less_than(self):
        self.run_test("1 < 2", True)
        self.run_test("1.0 < 2.0", True)
        self.run_test('"hello" < "world"', True)
        self.run_test("{1} < {2}", None)
        self.run_test("1 < 1", False)
        self.run_test('"hello" < "hello"', False)
        self.run_test('"hello" < 1', None)

    def test_less_equals(self):
        self.run_test("1 <= 2", True)
        self.run_test("1.0 <= 2.0", True)
        self.run_test('"hello" <= "world"', True)
        self.run_test("{1} <= {2}", None)
        self.run_test("1 <= 1", True)
        self.run_test('"hello" <= "hello"', True)
        self.run_test('"hello" <= 1', None)

    def test_and(self):
        self.run_test("true and true", True)
        self.run_test("true and false", False)
        self.run_test("false and true", False)
        self.run_test("false and false", False)
        self.run_test('"hello" and "world"', "world")
        self.run_test("{1} and {2}", [2])
        self.run_test("0 and {5}", [5])
        self.run_test("nil and {5}", NilVal())

    def test_or(self):
        self.run_test("true or true", True)
        self.run_test("true or false", True)
        self.run_test("false or true", True)
        self.run_test("false or false", False)
        self.run_test('"hello" or "world"', "hello")
        self.run_test("{1} or {2}", [1])
        self.run_test("0 or {5}", 0)
        self.run_test("nil or {5}", [5])


if __name__ == "__main__":
    unittest.main()



from tumfl.AST import ASTNode, Chunk
from tumfl.minifier.resolve_indexes import Resolve
from tumfl.minifier.shorten_names import GetNames, remove_unused_names
from tumfl.minifier.simplify_expressions import Simplify
from tumfl.minifier.util.replace_name import full_replace


def minify(ast: ASTNode) -> None:
    Resolve()(ast)
    r = GetNames()
    r(ast)
    remove_unused_names(ast, r)
    # print(r.current_scope)
    assert isinstance(ast, Chunk)
    replacements = r.current_scope.collect_replacements()
    full_replace(ast, replacements)
    # CombineLocal()(ast) combine_local still contains flaws
    Simplify()(ast)



from __future__ import annotations

from typing import Generic, TypeVar

from _weakrefset import WeakSet

V = TypeVar("V")


class SharedList(Generic[V]):
    # pylint: disable=protected-access
    def __init__(self, member: V):
        self._members: WeakSet[V] = WeakSet((member,))

    def absorb(self, other: SharedList) -> list[V]:
        self._members.update(other._members)
        for obj in list(other._members):
            obj._members = self._members
        return list(other._members)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from .expression import Expression

if TYPE_CHECKING:
    from tumfl.AST import BaseFunctionDefinition, Block, Name, Vararg
    from tumfl.Token import Token


class ExpFunctionDefinition(Expression):
    """Definition of a lua function expression"""

    def __init__(
        self,
        token: Token,
        parameters: Sequence[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "ExpFunctionDefinition")
        self.parameters: list[Name | Vararg] = list(parameters)
        self.body: Block = body

    @staticmethod
    def from_base_definition(
        base: BaseFunctionDefinition,
    ) -> ExpFunctionDefinition:
        return ExpFunctionDefinition(base.token, base.parameters, base.body)



from .bin_op import BinaryOperand, BinOp
from .boolean import Boolean
from .exp_function_call import ExpFunctionCall
from .exp_function_definition import ExpFunctionDefinition
from .exp_method_invocation import ExpMethodInvocation
from .expression import Expression
from .index import Index
from .name import Name
from .named_index import NamedIndex
from .nil import Nil
from .number import Number
from .string import String
from .table import Table
from .table_field import (
    ExplicitTableField,
    NamedTableField,
    NumberedTableField,
    TableField,
)
from .un_op import UnaryOperand, UnOp
from .vararg import Vararg
from .variable import Variable

__all__ = [
    "BinaryOperand",
    "BinOp",
    "Boolean",
    "ExpFunctionCall",
    "ExpFunctionDefinition",
    "ExpMethodInvocation",
    "Expression",
    "Index",
    "Name",
    "NamedIndex",
    "Nil",
    "Number",
    "String",
    "Table",
    "ExplicitTableField",
    "NamedTableField",
    "NumberedTableField",
    "TableField",
    "UnaryOperand",
    "UnOp",
    "Vararg",
    "Variable",
]



import tumfl.AST

from .dependency_resolver import resolve_recursive
from .error import LexerError, ParserError
from .formatter import format
from .minifier import minify
from .parser import parse
from .Token import Token

__all__ = [
    "resolve_recursive",
    "LexerError",
    "ParserError",
    "format",
    "minify",
    "parse",
    "Token",
]



from __future__ import annotations

import itertools
import string
from typing import TYPE_CHECKING, Iterator, Optional

from tumfl.AST import (
    Assign,
    ASTNode,
    Block,
    Expression,
    FunctionDefinition,
    MethodInvocation,
    Name,
    NamedIndex,
    Vararg,
)

if TYPE_CHECKING:
    from tumfl.minifier.util.replacements import ReplacementCollection, Replacements


def inner_add_alias(
    chunk: Block, current: Replacements, to_alias: list[Replacements], pos: list[int]
) -> None:
    if current.parent in to_alias:
        inner_add_alias(chunk, current.parent, to_alias, pos)
    if current in to_alias:
        to_alias.remove(current)
    assert current.replacement
    assert current.original
    token = current.original.token
    names: list[Name] = [current.original]
    todo: Name = names[0]
    while (
        isinstance(todo.parent_class, NamedIndex)
        and todo.parent_class.lhs is not names[0]
    ):
        if isinstance(todo.parent_class.lhs, NamedIndex):
            names.insert(0, todo.parent_class.lhs.variable_name)
        else:
            assert isinstance(todo.parent_class.lhs, Name)
            names.insert(0, todo.parent_class.lhs)
        todo = names[0]
    parent_idx: int = -1
    if current.parent is not None:
        for i, name in enumerate(names):
            if name in current.parent.targets:
                parent_idx = i
                break
        else:
            assert False, "Parent is set, but unable to find it in names"
    names = [name if i <= parent_idx else name.clone() for i, name in enumerate(names)]
    node: Expression = names.pop(0)
    for name in names:
        node = NamedIndex(token, node, name)
    assign = Assign(
        token,
        [Name(token, current.replacement)],
        [node],
    )
    if current.after is None:
        chunk.statements.insert(pos[0], assign)
        pos[0] += 1
    else:
        previous_assign: Optional[ASTNode] = current.after.parent_class
        assert isinstance(previous_assign, Assign)
        parent_block = previous_assign.parent_class
        assert isinstance(parent_block, Block)
        idx: int = parent_block.statements.index(previous_assign)
        assert idx >= 0
        parent_block.statements.insert(idx + 1, assign)


def add_aliases(chunk: Block, to_alias: list[Replacements]) -> None:
    pos = [0]
    while to_alias:
        replacement = to_alias.pop(0)
        inner_add_alias(chunk, replacement, to_alias, pos)


def replace_name(name: Name, replacement: str) -> None:
    parent: Optional[ASTNode] = name.parent_class
    name.variable_name = replacement
    assert parent
    if isinstance(parent, FunctionDefinition):
        idx: int
        other_name: Name | Vararg
        for i, other_name in enumerate(parent.names):
            if other_name is name:
                idx = i
                break
        else:
            for other_name in parent.parameters:
                if other_name is name:
                    return
            raise NotImplementedError("Name can not be method")
        parent.names[: idx + 1] = [name]
    elif isinstance(parent, MethodInvocation):
        if name is parent.method:
            raise NotImplementedError("Name can not be method")
    elif isinstance(parent, NamedIndex):
        if name is parent.variable_name:
            parent.replace(name)
        else:
            assert name is parent.lhs


def char_sequence() -> Iterator[str]:
    # Define the “alphabet” in order: a–z then A–Z
    alphabet = list(string.ascii_lowercase) + list(string.ascii_uppercase)
    length = 1
    while True:
        # For the current string-length `length`, yield all combinations in lexicographic order
        for tup in itertools.product(alphabet, repeat=length):
            yield "".join(tup)
        length += 1


def full_replace(chunk: Block, replacements: ReplacementCollection) -> None:
    for inner_loop, new_name in zip(replacements, char_sequence()):
        for replacement in inner_loop:
            replacement.replacement = new_name
    to_alias: list[Replacements] = []
    for inner_loop in replacements:
        for replacement in inner_loop:
            if replacement.original:
                to_alias.append(replacement)
    add_aliases(chunk, to_alias)
    for inner_loop in replacements:
        for replacement in inner_loop:
            assert replacement.replacement
            for name in replacement.targets:
                replace_name(name, replacement.replacement)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Expression
    from tumfl.Token import Token


class Assign(Statement):
    """The assignment statement, like a,b = 1,nil"""

    def __init__(
        self,
        token: Token,
        targets: Sequence[Expression],
        expressions: Sequence[Expression],
    ):
        super().__init__(token, "Assign")
        self.targets: list[Expression] = list(targets)
        self.expressions: list[Expression] = list(expressions)



from __future__ import annotations

from typing import Optional, Sequence

from tumfl.AST import (
    Assign,
    ASTNode,
    Block,
    ExpFunctionDefinition,
    ExplicitTableField,
    ExpMethodInvocation,
    FunctionDefinition,
    Index,
    IterativeFor,
    LocalAssign,
    LocalFunctionDefinition,
    MethodInvocation,
    Name,
    NamedIndex,
    NamedTableField,
    NumberedTableField,
    NumericFor,
    Statement,
    Table,
    Vararg,
)
from tumfl.basic_walker import NoneWalker
from tumfl.minifier.util.remove_name import RemoveName
from tumfl.minifier.util.scope import Scope
from tumfl.minifier.util.variable import Variable


class GetNames(NoneWalker):
    # pylint: disable=too-many-public-methods
    def __init__(self, ignored_names: Optional[Sequence[str]] = None) -> None:
        super().__init__()
        self.current_scope: Scope = Scope()
        self.scope_initialized: bool = False
        self.has_metatable: bool = False
        self.preserve: bool = False
        self.ignored_names: set[str] = set(ignored_names) if ignored_names else set()

    def push_outer_scope(self) -> None:
        self.scope_initialized = True
        self.push_scope()

    def push_scope(self) -> None:
        self.current_scope = Scope(parent=self.current_scope)

    def pop_scope(self) -> None:
        assert self.current_scope.parent, "Can not pop the root scope"
        self.current_scope = self.current_scope.parent

    def add_var_opt(self, node: ASTNode, global_: bool = False) -> None:
        if isinstance(node, Name) and node.variable_name not in self.ignored_names:
            self.current_scope.add_variable(node, self.preserve, global_)
        elif isinstance(node, NamedIndex):
            self.read_var(node.lhs)
            self.handle_subscript(node.lhs, node.variable_name, True)

    def read_var(self, node: ASTNode) -> None:
        if isinstance(node, Name) and node.variable_name not in self.ignored_names:
            self.current_scope.use_variable(node, False)
        elif isinstance(node, NamedIndex):
            self.read_var(node.lhs)
            self.handle_subscript(node.lhs, node.variable_name)

    def handle_subscript(
        self,
        base: ASTNode,
        extension: ASTNode,
        assign: bool = False,
        is_method: bool = False,
    ) -> None:
        if (var := base.get_attribute(Variable)) and isinstance(extension, Name):
            new_var = var.add_variable(extension)
            extension.set_attribute_if_not_exists(new_var)
            if assign:
                new_var.add_write(extension)
            else:
                new_var.add_read(extension, is_method)

    def set_preserve(self, node: Statement) -> None:
        if any("preserve" in comment for comment in node.token.comment):
            self.preserve = True
        elif any("preserve" in comment for comment in node.comment):
            self.preserve = True

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> None:
        self.set_preserve(node)
        last_name: Name = node.names[0]
        self.add_var_opt(last_name, global_=True)
        for name in node.names[1:]:
            self.handle_subscript(last_name, name, assign=True)
            last_name = name
        if node.method_name:
            self.handle_subscript(
                last_name, node.method_name, assign=True, is_method=True
            )
        self.preserve = False
        self.push_outer_scope()
        for param in node.parameters:
            self.add_var_opt(param)
        super().visit_FunctionDefinition(node)

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> None:
        self.add_var_opt(node.function_name)
        self.push_outer_scope()
        for param in node.parameters:
            self.add_var_opt(param)
        super().visit_LocalFunctionDefinition(node)

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> None:
        self.push_outer_scope()
        super().visit_ExpFunctionDefinition(node)

    def visit_Name(self, node: Name) -> None:
        if node.variable_name == "setmetatable":
            self.has_metatable = True
        assert node.parent_class
        parent: ASTNode = node.parent_class
        do_it: bool = (
            not isinstance(
                parent,
                (
                    NamedIndex,
                    NamedTableField,
                    ExpMethodInvocation,
                    MethodInvocation,
                    FunctionDefinition,
                    LocalFunctionDefinition,
                ),
            )
            or isinstance(parent, (ExpMethodInvocation, MethodInvocation))
            and node is not parent.method
            or isinstance(parent, NamedIndex)
            and node is not parent.variable_name
            or isinstance(parent, NamedTableField)
            and node is not parent.field_name
            or isinstance(parent, FunctionDefinition)
            and node is not parent.method_name
            and node not in parent.names
            or isinstance(parent, LocalFunctionDefinition)
            and node is not parent.function_name
        )
        if do_it:
            self.read_var(node)

    def visit_Assign(self, node: Assign) -> None:
        self.set_preserve(node)
        for target in node.targets:
            self.add_var_opt(target, True)
        self.preserve = False
        super().visit_Assign(node)

    def visit_LocalAssign(self, node: LocalAssign) -> None:
        for name in node.variable_names:
            self.add_var_opt(name.name)
        super().visit_LocalAssign(node)

    def visit_IterativeFor(self, node: IterativeFor) -> None:
        self.push_outer_scope()
        for name in node.namelist:
            self.add_var_opt(name)
        super().visit_IterativeFor(node)

    def visit_NumericFor(self, node: NumericFor) -> None:
        self.push_outer_scope()
        self.add_var_opt(node.variable_name)
        super().visit_NumericFor(node)

    def visit_Block(self, node: Block) -> None:
        if not self.scope_initialized:
            self.push_scope()
        self.scope_initialized = False
        super().visit_Block(node)
        self.pop_scope()

    def visit_NamedIndex(self, node: NamedIndex) -> None:
        super().visit_NamedIndex(node)
        self.handle_subscript(node.lhs, node.variable_name)

    def visit_NamedTableField(self, node: NamedTableField) -> None:
        assert node.parent_class
        self.handle_subscript(node.parent_class, node.field_name, True)
        super().visit_NamedTableField(node)

    def visit_ExplicitTableField(self, node: ExplicitTableField) -> None:
        super().visit_ExplicitTableField(node)
        assert node.parent_class
        if var := node.parent_class.get_attribute(Variable):
            var.has_variable_access.value = True

    def visit_NumberedTableField(self, node: NumberedTableField) -> None:
        super().visit_NumberedTableField(node)
        assert node.parent_class
        if var := node.parent_class.get_attribute(Variable):
            var.has_variable_access.value = True

    def visit_MethodInvocation(self, node: MethodInvocation) -> None:
        super().visit_MethodInvocation(node)
        self.handle_subscript(node.function, node.method, is_method=True)

    def visit_Index(self, node: Index) -> None:
        super().visit_Index(node)
        if var := node.lhs.get_attribute(Variable):
            var.has_variable_access.value = True

    def visit_Table(self, node: Table) -> None:
        if not node.get_attribute(Variable):
            node.set_attribute_if_not_exists(Variable(False, False))
        super().visit_Table(node)


class RemoveUnused(NoneWalker):
    def __init__(self) -> None:
        super().__init__()
        self.found_any: bool = False
        self.remove: RemoveName = RemoveName()

    def cleanup(self, node: ASTNode) -> bool:
        for comment in node.token.comment:
            if "tumfl" in comment and "preserve" in comment:
                return False
        self.found_any = True
        self.remove(node)
        if node.parent_class:
            node.remove()
        return True

    def visit_Assign(self, node: Assign) -> None:
        if all(
            isinstance(target, Name)
            and (var := target.get_attribute(Variable))
            and var.is_unused()
            for target in node.targets
        ):
            if self.cleanup(node):
                return
        super().visit_Assign(node)

    def visit_LocalAssign(self, node: LocalAssign) -> None:
        if all(
            (var := name.name.get_attribute(Variable)) and var.is_unused()
            for name in node.variable_names
        ):
            if self.cleanup(node):
                return
        super().visit_LocalAssign(node)

    def cleanup_params(self, params: list[Name | Vararg]) -> None:
        if all(not isinstance(param, Vararg) for param in params):
            for i in range(len(params) - 1, -1, -1):
                if (var := params[i].get_attribute(Variable)) and var.is_unused():
                    if self.cleanup(params[i]):
                        params.pop(i)
                    else:
                        break
                else:
                    break

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> None:
        if (var := node.names[0].get_attribute(Variable)) and var.is_unused():
            if self.cleanup(node):
                return
        self.cleanup_params(node.parameters)
        super().visit_FunctionDefinition(node)

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> None:
        if (var := node.function_name.get_attribute(Variable)) and var.is_unused():
            if self.cleanup(node):
                return
        self.cleanup_params(node.parameters)
        super().visit_LocalFunctionDefinition(node)

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> None:
        self.cleanup_params(node.parameters)
        super().visit_ExpFunctionDefinition(node)


def remove_unused_names(node: ASTNode, names: GetNames) -> None:
    remove = RemoveUnused()
    remove.found_any = True
    while remove.found_any:
        remove.found_any = False
        remove(node)
        names.current_scope.cleanup_scopes()



from __future__ import annotations

from tumfl.AST import Name
from tumfl.basic_walker import NoneWalker
from tumfl.minifier.util.variable import Variable


class RemoveName(NoneWalker):
    def visit_Name(self, node: Name) -> None:
        if var := node.get_attribute(Variable):
            var.remove(node)



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import Any, Generator, Optional, Type, TypeVar
from weakref import ReferenceType, ref

from tumfl.Token import Token
from tumfl.utils import generic_str

T = TypeVar("T")


class ASTNode(ABC):
    """Base class for all AST nodes"""

    def __init__(self, token: Token, name: str) -> None:
        self.name: str = name
        self.token: Token = token
        self._parent_class: Optional[ReferenceType[ASTNode]] = None
        self.file_name: Optional[Path] = None
        self.attributes: dict[Type[Any], Any] = {}

    @property
    def parent_class(self) -> Optional[ASTNode]:
        return None if self._parent_class is None else self._parent_class()

    @parent_class.setter
    def parent_class(self, value: Optional[ASTNode]) -> None:
        self._parent_class = ref(value) if value is not None else None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return all(
            callable(getattr(self, i)) or getattr(self, i) == getattr(other, i)
            for i in self.__dir()
        )

    def __repr__(self) -> str:
        return generic_str(self, ["parent_class"])

    def __dir(self) -> Generator[str, None, None]:
        return (
            i
            for i in dir(self)
            if not i.startswith("_")
            # ignore "token" for comparison (and parent check)
            and i
            not in [
                "parent_class",
                "token",
                "comment",
                "attributes",
                "file_name",
                "name",
            ]
        )

    def __get_children(self) -> Generator[ASTNode, None, None]:
        from tumfl.AST.statement.local_assign import AttributedName

        for name in self.__dir():
            node: Any = getattr(self, name)
            if isinstance(node, ASTNode):
                yield node
            elif isinstance(node, list):
                for child in node:
                    if isinstance(child, ASTNode):
                        yield child
                    elif isinstance(child, AttributedName):
                        yield child.name
                        if child.attribute is not None:
                            yield child.attribute

    def parent(
        self, parent: Optional[ASTNode], file_name: Optional[Path] = None
    ) -> None:

        self.parent_class = parent
        self.file_name = file_name
        for child in self.__get_children():
            child.parent(self, file_name)

    def replace(self, replacement: ASTNode) -> None:
        """Replaces node in-place"""
        self.__class__ = replacement.__class__  # type: ignore
        self.__dict__.clear()
        self.__dict__.update(replacement.__dict__)
        for child in self.__get_children():
            child.parent_class = self

    def remove(self) -> None:
        """Removes a child by replacing it with Semicolon or Boolean"""
        # Previously, this did remove the child node, but that leads
        # to issues when iterating over the children
        from tumfl.AST import Boolean, Semicolon, Statement

        replacement: ASTNode
        if isinstance(self, Statement):
            replacement = Semicolon(self.token)
        else:
            replacement = Boolean(self.token, False)
        replacement.parent(self.parent_class, self.file_name)
        self.replace(replacement)

    def set_attribute(self, value: Any) -> None:
        self.attributes[type(value)] = value

    def set_attribute_if_not_exists(self, value: Any) -> None:
        if type(value) not in self.attributes:
            self.set_attribute(value)

    def get_attribute(self, key: Type[T]) -> Optional[T]:
        return self.attributes.get(key)

    def __hash__(self) -> int:
        return hash(
            tuple(
                var
                for name in self.__dir()
                if not callable(var := getattr(self, name))
                and getattr(var, "__hash__", None)
            )
        )



from __future__ import annotations


class MutableBool:
    """This is simply a class that can be shared, and has a bool with interior mutability"""

    def __init__(self, value: bool):
        self.value = value

    def __bool__(self) -> bool:
        return self.value

    def __repr__(self) -> str:
        return f"MutableBool({self.value})"

    def __or__(self, other: bool | MutableBool) -> bool:
        self.value = self.value or bool(other)
        return self.value



import re

from tumfl.AST import (
    Boolean,
    ExplicitTableField,
    Expression,
    Name,
    NamedTableField,
    Nil,
    Number,
    NumberedTableField,
    String,
    Table,
    UnaryOperand,
    UnOp,
)
from tumfl.lexer import Lexer
from tumfl.to_python_type import NilVal, Retype
from tumfl.Token import Token, TokenType

NAME_REGEX: re.Pattern = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*$")


def to_lua_type(python_type: Retype) -> Expression:
    if isinstance(python_type, dict):
        table = Table(Token(TokenType.L_CURL, "{", 0, 0), [])
        i = 1
        try:
            while True:
                val = python_type.pop(i)
                lua_value = to_lua_type(val)
                table.fields.append(NumberedTableField(lua_value.token, lua_value))
                i += 1
        except KeyError:
            pass
        for key, val in python_type.items():
            lua_value = to_lua_type(val)
            if isinstance(key, str) and NAME_REGEX.fullmatch(key):
                name = Name.from_token(Token(TokenType.NAME, key, 0, 0))
                table.fields.append(NamedTableField(name.token, name, lua_value))
            else:
                key_value = to_lua_type(key)
                table.fields.append(
                    ExplicitTableField(key_value.token, key_value, lua_value)
                )
        return table
    if isinstance(python_type, list):
        table = Table(Token(TokenType.L_CURL, "{", 0, 0), [])
        for field in python_type:
            lua_value = to_lua_type(field)
            table.fields.append(NumberedTableField(lua_value.token, lua_value))
        return table
    if isinstance(python_type, bool):
        return Boolean.from_token(
            Token(
                TokenType.TRUE if python_type else TokenType.FALSE, python_type, 0, 0
            ),
        )
    if isinstance(python_type, (int, float)):
        number = str(python_type)
        signed = number.startswith("-")
        if signed:
            number = number[1:]
        lexed_number = Lexer(number).get_number()
        ast_number: Expression = Number.from_token(
            Token(TokenType.NUMBER, lexed_number, 0, 0)
        )
        if signed:
            ast_number = UnOp(
                Token(TokenType.MINUS, "-", 0, 0), UnaryOperand.MINUS, ast_number
            )
        return ast_number
    if isinstance(python_type, str):
        return String.from_token(Token(TokenType.STRING, python_type, 0, 0))
    if isinstance(python_type, NilVal):
        return Nil.from_token(Token(TokenType.NIL, "nil", 0, 0))
    raise NotImplementedError(f'Can not convert "{python_type}" to lua type')



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from tumfl.AST.ast_node import ASTNode

if TYPE_CHECKING:
    from tumfl.AST import Expression, Name
    from tumfl.Token import Token


class TableField(ASTNode, ABC):
    """Base class for all table fields"""

    def __init__(self, token: Token, name: str, value: Expression):
        super().__init__(token, name)
        self.value: Expression = value


class ExplicitTableField(TableField):
    """Table field of the form `[ expr ] = expr`"""

    def __init__(self, token: Token, at: Expression, value: Expression):
        super().__init__(token, "ExplicitTableField", value)
        self.at: Expression = at


class NamedTableField(TableField):
    """Table field of the form `Name = expr`"""

    def __init__(self, token: Token, field_name: Name, value: Expression):
        super().__init__(token, "NamedTableField", value)
        self.field_name: Name = field_name


class NumberedTableField(TableField):
    """Table field of the form `expr`"""

    def __init__(self, token: Token, value: Expression):
        super().__init__(token, "NumberedTableField", value)



import unittest
from pathlib import Path

from tumfl.AST import (
    Assign,
    Block,
    ExpFunctionCall,
    ExpFunctionDefinition,
    Number,
    Semicolon,
    String,
)
from tumfl.config import parse_config
from tumfl.dependency_resolver import (
    Chunk,
    FunctionCall,
    InvalidDependencyError,
    resolve_recursive,
)


class TestDependencyResolver(unittest.TestCase):
    def setUp(self) -> None:
        self.base_path: Path = Path(__file__).parent / "test_files"
        self.main_script: Path = self.base_path / "main.lua"
        assert self.main_script.is_file()

    def test_resolver(self) -> None:
        ast = resolve_recursive(self.main_script, [self.base_path])
        assert isinstance(ast, Chunk)
        self.assertEqual(len(ast.statements), 4)
        self.assertIsInstance(ast.statements[0], Chunk)
        self.assertIsInstance(ast.statements[1], Chunk)
        self.assertIsInstance(ast.statements[2], Semicolon)
        self.assertIsInstance(ast.statements[3], FunctionCall)

    def test_errors(self) -> None:
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "empty_path.lua", [self.base_path])
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "nonexistent_path.lua", [self.base_path])
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "number_path.lua", [self.base_path])

    def test_config(self):
        config = parse_config(self.base_path / "config.lua")

        ast = resolve_recursive(
            self.base_path / "config_test.lua", [self.base_path], config=config
        )
        assert isinstance(ast, Chunk)
        self.assertEqual(len(ast.statements), 1)
        assign = ast.statements[0]
        self.assertIsInstance(assign, Assign)
        assert isinstance(assign, Assign)
        self.assertEqual(len(assign.expressions), 2)
        first_value = assign.expressions[0]
        assert isinstance(first_value, String)
        self.assertEqual(first_value.value, "foo")
        second_value = assign.expressions[1]
        assert isinstance(second_value, Number)
        self.assertEqual(second_value.to_int(), 42)

    def test_add_source_description(self):
        ast = resolve_recursive(
            self.main_script, [self.base_path], add_source_description=True
        )
        assert isinstance(ast, Chunk)
        first_statement = ast.statements[0]
        self.assertEqual(len(first_statement.comment), 1)
        self.assertIn(
            "Sourced from",
            first_statement.comment[0],
        )
        self.assertIn(
            str(self.base_path / "a.lua"),
            first_statement.comment[0],
        )

    def test_multi_import(self):
        ast = resolve_recursive(self.base_path / "multi_import.lua", [self.base_path])
        assert isinstance(ast, Chunk)
        self.assertEqual(len(ast.statements), 1)
        first_statement = ast.statements[0]
        self.assertIsInstance(first_statement, Block)

    def test_invalid_path(self):
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(
                self.base_path / "invalid_require_path.lua", [self.base_path]
            )
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(
                self.base_path / "invalid_require_type.lua", [self.base_path]
            )

    def test_require_expr(self):
        ast = resolve_recursive(self.base_path / "require_expr.lua", [self.base_path])
        assert isinstance(ast, Chunk)
        self.assertEqual(len(ast.statements), 1)
        first_statement = ast.statements[0]
        self.assertIsInstance(first_statement, Assign)
        assert isinstance(first_statement, Assign)
        expression = first_statement.expressions[0]
        self.assertIsInstance(expression, ExpFunctionCall)
        assert isinstance(expression, ExpFunctionCall)
        self.assertEqual(len(expression.arguments), 0)
        func_def = expression.function
        self.assertIsInstance(func_def, ExpFunctionDefinition)
        assert isinstance(func_def, ExpFunctionDefinition)
        self.assertEqual(len(func_def.parameters), 0)
        self.assertEqual(len(func_def.body.statements), 0)
        assert func_def.body.returns is not None
        self.assertEqual(len(func_def.body.returns), 1)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from .statement import Statement

if TYPE_CHECKING:
    from tumfl.AST import Block, Expression, Name
    from tumfl.Token import Token


class IterativeFor(Statement):
    """The iterative for, i.e. using pairs() or ipairs()"""

    def __init__(
        self,
        token: Token,
        namelist: Sequence[Name],
        explist: Sequence[Expression],
        body: Block,
    ):
        super().__init__(token, "IterativeFor")
        self.namelist: list[Name] = list(namelist)
        self.explist: list[Expression] = list(explist)
        self.body: Block = body



from .assign import Assign
from .block import Block
from .break_ import Break
from .chunk import Chunk
from .function_call import FunctionCall
from .function_definition import FunctionDefinition
from .goto import Goto
from .if_ import If
from .iterative_for import IterativeFor
from .label import Label
from .local_assign import AttributedName, LocalAssign
from .local_function_definition import LocalFunctionDefinition
from .method_invocation import MethodInvocation
from .numeric_for import NumericFor
from .repeat import Repeat
from .semicolon import Semicolon
from .statement import Statement
from .while_ import While

__all__ = [
    "Assign",
    "Block",
    "Break",
    "Chunk",
    "FunctionCall",
    "FunctionDefinition",
    "Goto",
    "If",
    "IterativeFor",
    "Label",
    "AttributedName",
    "LocalAssign",
    "LocalFunctionDefinition",
    "MethodInvocation",
    "NumericFor",
    "Repeat",
    "Semicolon",
    "Statement",
    "While",
]



import re

from tumfl.basic_walker import (
    ExplicitTableField,
    Index,
    Name,
    NamedIndex,
    NamedTableField,
    NoneWalker,
    String,
)
from tumfl.lexer import RESERVED_KEYWORDS

VALID_NAME: re.Pattern[str] = re.compile(r"[a-zA-Z_][a-zA-Z_0-9]*")


def is_valid_name(name: str) -> bool:
    return bool(VALID_NAME.fullmatch(name)) and name not in RESERVED_KEYWORDS


class Resolve(NoneWalker):
    """
    replaces `foo["bar"]` with `foo.bar` and {"foo"=...} with {foo=...}
    This is mostly done in order to simplify further minifications
    """

    def visit_ExplicitTableField(self, node: ExplicitTableField) -> None:
        if isinstance(node.at, String) and is_valid_name(node.at.name):
            new_node = NamedTableField(
                node.token, Name(node.at.token, node.at.value), node.value
            )
            node.replace(new_node)
        else:
            super().visit_ExplicitTableField(node)

    def visit_Index(self, node: Index) -> None:
        if isinstance(node.variable_name, String) and is_valid_name(
            node.variable_name.value
        ):
            new_node = NamedIndex(
                node.token,
                node.lhs,
                Name(node.variable_name.token, node.variable_name.value),
            )
            node.replace(new_node)
        else:
            super().visit_Index(node)
